using System;
using System.Drawing;
using System.Windows.Forms;
using System.Collections.Generic;
using SlimDX;
using SlimDX.DirectInput;

namespace PolledDevices
{
    /// <summary>
    /// 
    /// </summary>
	public class PolledJoystick : PolledDevice
    {
        #region Constants

        private const int JoystickMin = -100;
		private const int JoystickMax = +100;

        #endregion

        #region Fields

        private JoystickState _state;
		private int _positionThreshold;
		private int _rotationThreshold;
		private int _sliderThreshold;

        #endregion

        #region Events

        public event PolledJoystickEventHandler JoystickPolledEvent;
		public event PolledJoystickEventHandler JoystickPositionEvent;
		public event PolledJoystickEventHandler JoystickRotationEvent;
		public event PolledJoystickEventHandler JoystickSliderEvent;
		public event PolledJoystickEventHandler JoystickPointOfViewEvent;
		public event PolledJoystickEventHandler JoystickDownEvent;
		public event PolledJoystickEventHandler JoystickUpEvent;

        #endregion

        #region Constructor

        /// <summary>
        /// 
        /// </summary>
        /// <param name="owner"></param>
        public PolledJoystick(Control owner) : base(owner)
		{
			_positionThreshold = 1;
			_rotationThreshold = 1;
			_sliderThreshold = 1;
		}

		#endregion

		#region PolledDevice Implementations

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
		protected override bool AcquireDevice()
		{
            // enumerate through all the game controllers on the system and pick the first one
            IList<DeviceInstance> dilist = _dinput.GetDevices(DeviceClass.GameController, DeviceEnumerationFlags.AttachedOnly);
            if (dilist != null && dilist.Count > 0)
            {
                IEnumerator<DeviceInstance> di = dilist.GetEnumerator();
                while (di.MoveNext())
                {
                    _device = new Joystick(_dinput, di.Current.InstanceGuid);
                    break; // just take the first attached joystick
                }
            }
            else return false; // no joysticks

			_device.SetCooperativeLevel(_owner, CooperativeLevel.Background | CooperativeLevel.Nonexclusive);
            _device.Properties.AxisMode = DeviceAxisMode.Absolute;

			// set all of the axes on the controller to range (JoystickMin, JoystickMax)
			IList<DeviceObjectInstance> doilist = _device.GetObjects();
            if (doilist != null && doilist.Count > 0)
            {
                IEnumerator<DeviceObjectInstance> doi = doilist.GetEnumerator();
                while (doi.MoveNext())
                {
                    if ((doi.Current.ObjectType & ObjectDeviceType.Axis) != 0)
                    {
                        _device.Properties.SetRange(JoystickMin, JoystickMax); // default
                        _device.Properties.DeadZone = 0; // default
                    }
                }
            }
			_device.Acquire();
            _state = ((Joystick) _device).GetCurrentState();
			return true;
		}

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
		protected override bool CompareState()
		{
			JoystickState newState;
			try
			{
                newState = ((Joystick) _device).GetCurrentState();
			}
            catch (SlimDX.DirectInput.DirectInputException)
			{
				return false;
			}

			// position (x, y, z axes)
			bool positionChanged = 
                   Math.Abs(_state.X - newState.X) >= _positionThreshold 
				|| Math.Abs(_state.Y - newState.Y) >= _positionThreshold 
				|| Math.Abs(_state.Z - newState.Z) >= _positionThreshold;

			// rotation (Rx, Ry, Rz axes)
			bool rotationChanged =
                   Math.Abs(_state.RotationX - newState.RotationX) >= _rotationThreshold
                || Math.Abs(_state.RotationY - newState.RotationY) >= _rotationThreshold
                || Math.Abs(_state.RotationZ - newState.RotationZ) >= _rotationThreshold;

			// sliders (u-axis and v-axis)
			int[] oldSliders = _state.GetSliders();
			int[] newSliders = newState.GetSliders();
			bool sliderChanged = 
                   Math.Abs(oldSliders[0] - newSliders[0]) >= _sliderThreshold 
				|| Math.Abs(oldSliders[1] - newSliders[1]) >= _sliderThreshold;

			// point-of-view
            int[] oldPOVs = this._state.GetPointOfViewControllers();
            int[] newPOVs = newState.GetPointOfViewControllers();
			bool povChanged = (oldPOVs[0] != newPOVs[0]);

			// get the button arrays
            bool[] oldButtons = _state.GetButtons();
            bool[] newButtons = newState.GetButtons();

            // copy the state reference before firing events
			this._state = newState;

			//
			// fire events henceforth!!
			//

			// fire generic updated event
			if (JoystickPolledEvent != null)
				JoystickPolledEvent(this, new PolledJoystickEventArgs(_state));

			// fire position event
			if (positionChanged && JoystickPositionEvent != null)
				JoystickPositionEvent(this, new PolledJoystickEventArgs(_state));

			// fire rotation event
			if (rotationChanged && JoystickRotationEvent != null)
				JoystickRotationEvent(this, new PolledJoystickEventArgs(_state));

			// fire the slider events
			if (sliderChanged && JoystickSliderEvent != null)
				JoystickSliderEvent(this, new PolledJoystickEventArgs(_state));

			// fire point-of-view events
			if (povChanged && JoystickPointOfViewEvent != null)
				JoystickPointOfViewEvent(this, new PolledJoystickEventArgs(_state));

			// fire button events
			if (JoystickDownEvent != null || JoystickUpEvent != null)
			{
				for (int i = 0; i < oldButtons.Length; i++)
				{
                    if (!oldButtons[i] && newButtons[i]) // button down
					{
						if (JoystickDownEvent != null)
							JoystickDownEvent(this, new PolledJoystickEventArgs(_state, i));
					}
					else if (oldButtons[i] && !newButtons[i]) // button up
					{
						if (JoystickUpEvent != null)
							JoystickUpEvent(this, new PolledJoystickEventArgs(_state, i));
					}
				}
			}

			return true;
		}

		#endregion

		#region Other Properties

        /// <summary>
        /// 
        /// </summary>
		public Rectangle CoordinatePlane
		{
			get
			{
				return new Rectangle(
					JoystickMin,				// x
					JoystickMin,				// y
					JoystickMax - JoystickMin,	// width
					JoystickMax - JoystickMin	// height
					);
			}
		}

        /// <summary>
        /// 0 to 10,000 == 0% to 100% from center
        /// </summary>
		public float DeadzonePercent
		{
			set
			{
				if (value < 0f)
					value = 0f;
				else if (value > 1f)
					value = 1f;
				
                IList<DeviceObjectInstance> doilist = _device.GetObjects();
                if (doilist != null && doilist.Count > 0)
                {
                    IEnumerator<DeviceObjectInstance> doi = doilist.GetEnumerator();
                    while (doi.MoveNext())
                    {
                        if ((doi.Current.ObjectType & ObjectDeviceType.Axis) != 0)
                        {
                            _device.Properties.DeadZone = (int) (value * 10000f);
                        }
                    }
                }
			}
			get
			{
                IList<DeviceObjectInstance> doilist = _device.GetObjects();
                if (doilist != null && doilist.Count > 0)
                {
                    IEnumerator<DeviceObjectInstance> doi = doilist.GetEnumerator();
                    while (doi.MoveNext())
                    {
                        if ((doi.Current.ObjectType & ObjectDeviceType.Axis) != 0)
                        {
                            return _device.Properties.DeadZone / 10000f; // return deadzone of first stick
                        }
                    }
                }
                return -1f;
			}
		}

        /// <summary>
        /// 
        /// </summary>
		public int PositionThreshold
		{
			get 
			{
				return _positionThreshold;
			}
			set
			{
				if (value <= 0)
					value = 1;
				_positionThreshold = value;
			}
		}

        /// <summary>
        /// 
        /// </summary>
		public int RotationThreshold
		{
			get
			{
				return _rotationThreshold;
			}
			set
			{
				if (value <= 0)
					value = 1;
				_rotationThreshold = value;
			}
		}

        /// <summary>
        /// 
        /// </summary>
		public int SliderThreshold
		{
			get
			{
				return _sliderThreshold;
			}
			set
			{
				if (value <= 0)
					value = 1;
				_sliderThreshold = value;
			}
		}

		#endregion
	}
}
